/* io.c : entrees/sorties bufferisees */

#include <fcntl.h>
#include "mlvalues.h"
#include "io.h"
#include "fail.h"
#include "alloc.h"
#include "signals.h"

/* Fonctions communes entree/sortie */

value open_descriptor(fd)       /* ML */
     value fd;
{
  struct channel * channel;

  channel = (struct channel *) stat_alloc(sizeof(struct channel));
  channel->fd = CINT(fd);
  channel->curr = channel->max = channel->buff;
  channel->end = channel->buff + IO_BUFFER_SIZE;
  return (value) channel;
}

value channel_descriptor(channel)   /* ML */
     struct channel * channel;
{
  return MLINT(channel->fd);
}

#ifdef ANSI
extern long lseek(int, long, int);
#else
extern long lseek();
#endif

value channel_size(channel)      /* ML */
     struct channel * channel;
{
  long start, end;

  start = lseek(channel->fd, 0L, 1);
  if (start == -1) os_error();
  end = lseek(channel->fd, 0L, 2);
  if (end == -1) os_error();
  if (lseek(channel->fd, start, 0) != start) os_error();
  return MLINT(end);
}

/******** Sorties *********/

static void really_write(fd, p, n)
     int fd;
     char * p;
     int n;
{
  int retcode;
  while (n > 0) {
    retcode = write(fd, p, n);
    if (retcode == -1) os_error();
    p += retcode;
    n -= retcode;
  }
}   

value flush(channel)            /* ML */
     struct channel * channel;
{
  int n;
  n = channel->max - channel->buff;
  if (n > 0)
    really_write(channel->fd, channel->buff, n);
  channel->curr = channel->max = channel->buff;
  return atom(0);
}

value output_char(channel, ch)  /* ML */
     struct channel * channel;
     value ch;
{
  putc(channel, CINT(ch));
  return atom(0);
}

void putw(channel, w)
     struct channel * channel;
     long w;
{
  putc(channel, w >> 24);
  putc(channel, w >> 16);
  putc(channel, w >> 8);
  putc(channel, w);
}

value output_int(channel, w)    /* ML */
     struct channel * channel;
     value w;
{
  putw(channel, CINT(w));
  return atom(0);
}

value output(channel, buff, start, length) /* ML */
     struct channel * channel;
     value buff, start, length;
{
  unsigned char * p;
  int n, m;

  p = &BYTE(buff, CINT(start));
  n = CINT(length);
  m = channel->end - channel->curr;
  if (channel->curr == channel->buff && n >= m) {
    really_write(channel->fd, p, n);
  } else if (n <= m) {
    bcopy(p, channel->curr, n);
    channel->curr += n;
    if (channel->curr > channel->max) channel->max = channel->curr;
  } else {
    bcopy(p, channel->curr, m);
    p += m;
    n -= m;
    m = channel->end - channel->buff;
    really_write(channel->fd, channel->buff, m);
    if (n <= m) {
      bcopy(p, channel->buff, n);
      channel->curr = channel->max = channel->buff + n;
    } else {
      really_write(channel->fd, p, n);
      channel->curr = channel->max = channel->buff;
    }
  }
  return atom(0);
}

value seek_out(channel, pos)    /* ML */
     struct channel * channel;
     value pos;
{
  long start, dest;

  dest = CINT(pos);
  start = lseek(channel->fd, 0, 1);
  if (start == -1) os_error();
  if (dest >= start && dest <= start + channel->end - channel->buff) {
    channel->curr = channel->buff + dest - start;
    if (channel->curr > channel->max) channel->max = channel->curr;
  } else {
    flush(channel);
    if (lseek(channel->fd, dest, 0) != dest) os_error();
  }
  return atom(0);
}

value pos_out(channel)          /* ML */
     struct channel * channel;
{
  long start;

  start = lseek(channel->fd, 0, 1);
  if (start == -1) os_error();
  return MLINT(start + channel->curr - channel->buff);
}

value close_out(channel)     /* ML */
     struct channel * channel;
{
  flush(channel);
  if (close(channel->fd) == -1) os_error();
  stat_free(channel);
  return atom(0);
}

/* Entrees */

static int really_read(fd, p, n)
     int fd;
     char * p;
     int n;
{
  int retcode;

  raise_on_signal = 1;          /* Pour ne pas rester bloque dans le read */
  if (signal_is_pending)        /* Si un signal s'est declenche entre */
    raise_pending_signal();     /* le dernier STARTFUN et raise_on_signal=1 */
  retcode = read(fd, p, n);
  raise_on_signal = 0;
  if (retcode == -1) os_error();
  return retcode;
}

unsigned char refill(channel)
     struct channel * channel;
{
  int n;

  n = really_read(channel->fd, channel->buff, IO_BUFFER_SIZE);
  if (n == 0) mlraise(atom(END_OF_FILE_EXN));
  channel->max = channel->buff + n;
  channel->curr = channel->buff + 1;
  return channel->buff[0];
}

value input_char(channel)       /* ML */
     struct channel * channel;
{
  char c;
  c = getc(channel);
  return MLINT(c);
}

long getw(channel)
     struct channel * channel;
{
  int i;
  long res;

  res = 0;
  for(i = 0; i < 4; i++)
    res = (res << 8) + getc(channel);
  return res;
}

value input_int(channel)        /* ML */
     struct channel * channel;
{
  return MLINT(getw(channel));
}

value input(channel, buff, start, length) /* ML */
     struct channel * channel;
     value buff, start, length;
{
  unsigned char * p;
  int m, n, l;

  p = &BYTE(buff, CINT(start));
  n = CINT(length);
  m = channel->max - channel->curr;
  if (n <= m) {
    bcopy(channel->curr, p, n);
    channel->curr += n;
    return MLINT(n);
  } else {
    bcopy(channel->curr, p, m);
    p += m;
    n -= m;
    if (n < IO_BUFFER_SIZE) {
      l = really_read(channel->fd, channel->buff, IO_BUFFER_SIZE);
      channel->max = channel->buff + l;
      if (n > l) n = l;
      bcopy(channel->buff, p, n);
      channel->curr = channel->buff + n;
      return MLINT(m + n);
    } else {
      channel->max = channel->curr = channel->buff;
      return MLINT(m + really_read(channel->fd, p, n));
    }
  }
}

value seek_in(channel, pos)     /* ML */
     struct channel * channel;
     value pos;
{
  long start, dest;

  dest = CINT(pos);
  start = lseek(channel->fd, 0, 1);
  if (start == -1) os_error();
  start -= channel->max - channel->buff;
  if (dest >= start && dest <= start + channel->max - channel->buff) {
    channel->curr = channel->buff + dest - start;
  } else {
    if (lseek(channel->fd, dest, 0) != dest) os_error();
    channel->curr = channel->max = channel->buff;
  }
  return atom(0);
}

value pos_in(channel)           /* ML */
     struct channel * channel;
{
  long start;

  start = lseek(channel->fd, 0, 1);
  if (start == -1) os_error();
  return MLINT(start - (channel->max - channel->curr));
}

value close_in(channel)     /* ML */
     struct channel * channel;
{
  if (close(channel->fd) == -1) os_error();
  stat_free(channel);
  return atom(0);
}

