(* parse_engine.ml : le moteur d'analyse syntaxique. *)

#open "obj";;
#open "meta";;
#open "lex_run";;
#open "lex_aux";;

let YY_FLAG = -1000;;

(* Taille des piles du parser *)

let yy_max_depth = ref 150;;

(* Valeur d'initialisation de la pile de valeurs *)

let v_stack_marker = Repr 0;;

(* Les piles: etats, valeurs, positions de debut, positions de fin *)

let s_stack = ref (make_vect (!yy_max_depth+1) 0)
and v_stack = ref (make_vect (!yy_max_depth+1) v_stack_marker)
and symbol_start_stack = ref (make_vect (!yy_max_depth+1) 0)
and symbol_end_stack = ref (make_vect (!yy_max_depth+1) 0)
;;

(* Niveau de production (dans les actions) *)
let yy_pvt = ref 0;;

(* Pour liberer les objets de la pile de valeurs *)

let clear_v_stack () =
  for i = 0 to vect_length !v_stack - 1 do
    (!v_stack).(i) <- v_stack_marker
  done
;;

(* En cas de debordement des piles *)

let grow_yacc_stacks () =
    let old_yy_max_depth = !yy_max_depth in
    yy_max_depth := 2*!yy_max_depth;
    let new_s_stack = make_vect (!yy_max_depth+1) 0
    and new_v_stack = make_vect (!yy_max_depth+1) v_stack_marker
    and new_symbol_start_stack = make_vect (!yy_max_depth+1) 0
    and new_symbol_end_stack = make_vect (!yy_max_depth+1) 0
    in
      for i = 0 to old_yy_max_depth do
        new_s_stack.(i) <- (!s_stack).(i);
        new_v_stack.(i) <- (!v_stack).(i);
        new_symbol_start_stack.(i) <- (!symbol_start_stack).(i);
        new_symbol_end_stack.(i) <- (!symbol_end_stack).(i)
      done;
    s_stack := new_s_stack;
    v_stack := new_v_stack;
    symbol_start_stack := new_symbol_start_stack;
    symbol_end_stack := new_symbol_end_stack;
    ()
;;

(* Acces a la valeur du nieme symbole, au cours d'une reduction *)

let peek_val n = magic_obj (!v_stack).(!yy_pvt - n);;

(* Variables d'etat du parser *)

let yy_val = ref (Repr 0)
and yy_symbol_start = ref 0
and yy_symbol_end = ref 0
and yy_pv = ref (-1)
;;

let symbol_start () = !yy_symbol_start
and symbol_end () = !yy_symbol_end
;;

let yy_char = ref (-1);;

exception yy_exit of obj;;              (* obj = la valeur semantique *)
exception Parsing of int;;              (* int = le token courant *)

let yy_parse
    (actions,
     (exca,act,pact,pgo,r1,r2,chk,def),
     YY_LAST)
    yy_lex =
  let yy_actions = actions in
  let yy_exca = exca in
  let yy_act = act in
  let yy_pact = pact in
  let yy_pgo = pgo in
  let yy_r1 = r1 in
  let yy_r2 = r2 in
  let yy_chk = chk in
  let yy_def = def in

  let rec stack yy_state =
  (* Stockage d'un etat et de ses attributs sur la pile *)
     if incr yy_pv > !yy_max_depth then grow_yacc_stacks();
(***
     print_string "State "; print_int yy_state;
     print_string ", char "; print_int !yy_char; print_newline(); 
     flush std_out;
***)
     (!s_stack).(!yy_pv) <- yy_state;
     (!v_stack).(!yy_pv) <- !yy_val;
     (!symbol_start_stack).(!yy_pv) <- !yy_symbol_start;
     (!symbol_end_stack).(!yy_pv) <- !yy_symbol_end;
     let yy_n1 = yy_pact.(yy_state) in
       if yy_n1 <= YY_FLAG then default yy_state else
        begin
         (* On lit un token, si c'est pas deja fait *)
         if !yy_char < 0 then
           begin yy_char := yy_lex (); () end;
         (* On essaye un shift *)  
         let yy_n2 = yy_n1 + !yy_char in
           if yy_n2 < 0 or yy_n2 >= YY_LAST then default yy_state else
             let yy_n3 = yy_act.(yy_n2) in
               if yy_chk.(yy_n3) != !yy_char then default yy_state else
                (* Shift valide *)
                begin
                 yy_char := -1;
                 yy_val := !lex_val;
                 lex_val := Repr 0;
                 yy_symbol_start := start_token();
                 yy_symbol_end := end_token();
                 stack yy_n3
                end
        end

  and default yy_state =
    let yy_n1 = yy_def.(yy_state) in
    let yy_n =
      if yy_n1 != -2 then yy_n1 else
       begin
        (* Recherche dans la table des exceptions *)
        if !yy_char < 0 then begin yy_char := yy_lex(); () end;
        let yy_n2 = search_exception1 yy_state 0 in
        if yy_n2 < 0 then raise (yy_exit !yy_val) else yy_n2
       end
    in
      if yy_n == 0 then raise (Parsing !yy_char) else
       begin
(***
        print_string "Reduce "; print_int yy_n; print_newline();
        flush std_out;
***)
        yy_pvt := !yy_pv;
        yy_pv := !yy_pv - yy_r2.(yy_n);
        yy_symbol_start := (!symbol_start_stack).(succ !yy_pv);
        yy_symbol_end := (!symbol_end_stack).(!yy_pvt);

        let yy_m = yy_r1.(yy_n) in
        let yy_j = succ(yy_pgo.(yy_m) + (!s_stack).(!yy_pv)) in

          (* On effectue l'action semantique associee a la regle *)
          yy_val := yy_actions yy_n;

          (* On consulte la table goto, pour trouver le prochain etat *)
            stack (if yy_j >= YY_LAST then yy_act.(yy_pgo.(yy_m))
                   else
                     let yy_default_state = yy_act.(yy_j) in
                       if yy_chk.(yy_default_state) == - yy_m
                       then yy_default_state
                       else yy_act.(yy_pgo.(yy_m)))
       end

  and search_exception1 yy_state i =
    if (yy_exca.(i) != -1) or (yy_exca.(succ i) != yy_state)
    then search_exception1 yy_state (i+2)
    else search_exception2 (i+2)
  and search_exception2 i =
    if (yy_exca.(i) >= 0) & (yy_exca.(i) != !yy_char)
    then search_exception2 (i+2)
    else yy_exca.(succ i)

  in
    fun initial_token ->
      yy_char := initial_token;
      try stack 0 with yy_exit v -> v
;;

let yy_exit_from_external v = raise (yy_exit v)
;;
